from .proxy import LineageRegistry
from .proxy import REGISTRY_NAME
from .testing import LINEAGEREGISTRY_INTEGRATION_TESTING
from .utils import disableRegistry
from .utils import enableRegistry
from Products.CMFCore.utils import getToolByName
from five.localsitemanager import make_objectmanager_site
from interlude import interact
from plone.app.registry.registry import Registry
from plone.app.testing import TEST_USER_ID
from plone.app.testing import setRoles
from plone.registry.interfaces import IRegistry
from plone.testing import layered
from plone.testing import z2
from zope import schema
from zope.component import getUtility
from zope.component.hooks import setSite
from zope.component.interfaces import ISite
from zope.interface import Interface

import doctest
import pprint
import unittest


TESTFILES = [
    ('proxy.rst', LINEAGEREGISTRY_INTEGRATION_TESTING),
]

optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
optionflags |= doctest.REPORT_ONLY_FIRST_FAILURE


class ITestSchema(Interface):
    """Test schema to test the registry.forInterface call.
    """
    test_attribute = schema.TextLine(default=u"test value")


class TestLineageRegistry(unittest.TestCase):
    layer = LINEAGEREGISTRY_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer['portal']

    def test_registry_assignment(self):
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        self.oid = self.portal.invokeFactory('Folder', 'folder')
        setRoles(self.portal, TEST_USER_ID, ['Member'])

        folder = self.portal[self.oid]
        registry = getUtility(IRegistry)

        if not ISite.providedBy(folder):
            make_objectmanager_site(folder)

        setSite(folder)

        pc = getToolByName(folder, 'portal_catalog')
        pc.reindexObject(folder, idxs=['object_provides'])

        enableRegistry(folder)
        self.assertIn(REGISTRY_NAME, folder.objectIds())

        registry = getUtility(IRegistry)
        self.assertTrue(isinstance(registry, LineageRegistry))

        disableRegistry(folder)
        registry = getUtility(IRegistry)
        self.assertTrue(isinstance(registry, Registry))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests([
        layered(
            doctest.DocFileSuite(
                docfile,
                globs={'interact': interact,
                       'pprint': pprint.pprint,
                       'z2': z2,
                       },
                optionflags=optionflags,
            ),
            layer=layer,
        )
        for docfile, layer in TESTFILES
    ])
    suite.addTest(unittest.makeSuite(TestLineageRegistry))
    return suite



from Acquisition import aq_parent
from persistent import Persistent
from plone.app.registry import Registry
from plone.registry import FieldRef
from plone.registry import Record
from plone.registry.interfaces import IRegistry
from plone.registry.registry import _Records
from zope.component import getSiteManager

_MARKER = object()

REGISTRY_NAME = 'lineage_registry'


def get_parent_registry(base_registry):
    # aq_parent returns the container, where the base_registry is contained
    parent = aq_parent(base_registry)

    # get the parent's site manager
    # If parent is None for pre lineage.registry 1.1 sites, return the current
    # site.
    sm = getSiteManager(context=parent)

    for base in sm._getBases():
        registry = base.queryUtility(IRegistry)
        if registry is not None:
            return registry
    return None


class LineageRegistry(Registry):
    """Registry which proxies values from its parent registry in the local
    component stack. On set it sets values in this layered registry.
    """

    def __init__(self, id, title=None, parent=None):
        if parent:
            # We have to acquisition-wrap the registry before the
            # _LineageRecords object is set, so that it's __parent__ attribute
            # is also set to the acquisition-wrapped registry.
            self = self.__of__(parent)
        else:
            raise AssertionError(
                "parent must not be None, otherwise the Acquisition context "
                "of LineageRegistry cannot be set")

        super(LineageRegistry, self).__init__(id, title)
        # pass the acquisition-wrapped LineageRegistry instance to
        # _LineageRecords
        self._records = _LineageRecords(self)

    def __getitem__(self, name):
        records = self._get_and_fix_records()
        return records[name].value

    def get(self, name, default=None):
        records = self._get_and_fix_records()
        record = records.get(name, _MARKER)
        if record is _MARKER:
            return default
        return record.value

    def __setitem__(self, name, value):
        records = self._get_and_fix_records()
        if name in records._values:
            records[name].value = value
            return
        record = Record(records[name].field, value)
        records[name] = record

    def __contains__(self, name):
        records = self._get_and_fix_records()
        return name in records

    def _get_and_fix_records(self):
        recs = self._records
        if not hasattr(recs.__parent__, '__parent__'):
            if not hasattr(self, '__parent__'):
                raise AttributeError(
                    "We lost the Acquisition context. This shouldn't happen.")
            recs.__parent__ = self
        return recs

    @property
    def records(self):
        return self._get_and_fix_records()


class _LineageRecords(_Records, Persistent):
    """The records stored in the registry. This implements dict-like access
    to records, where as the Registry object implements dict-like read-only
    access to values.
    """

    @property
    def _parent_records(self):
        registry = get_parent_registry(self.__parent__)
        if hasattr(registry, '_get_and_fix_records'):
            # LineageRegistry
            records = registry._get_and_fix_records()
        else:
            # Normal Registry
            records = registry.records
        return records

    def __setitem__(self, name, record):
        parent_rec = self._parent_records.get(name, _MARKER)
        if parent_rec is not _MARKER and parent_rec.value == record.value:
            return
        super(_LineageRecords, self).__setitem__(name, record)

    def __delitem__(self, name):
        if name not in self._values:
            return
        super(_LineageRecords, self).__delitem__(name)

    def __getitem__(self, name):
        if name not in self._values:
            return self._parent_records.__getitem__(name)
        return super(_LineageRecords, self).__getitem__(name)

    def __iter__(self):
        for name in self._values.__iter__():
            yield name
        for name in self._parent_records.__iter__():
            if name not in self._values:
                yield name

    def has_key(self, name):
        return name in self._values\
            or name in self._parent_records

    def __contains__(self, name):
        return self._values.__contains__(name)\
            or self._parent_records.__contains__(name)

    def keys(self, min=None, max=None):
        data = set(self._values.keys(min, max))
        data.update(self._parent_records.keys(min, max))
        return list(data)

    def maxKey(self, key=None):
        keys = []
        keys.append(self._parent_records.maxKey(key))
        try:
            own = self._values.maxKey(key)
            keys.append(own)
        except ValueError:
            pass
        return max(keys)

    def minKey(self, key=None):
        keys = []
        keys.append(self._parent_records.minKey(key))
        try:
            own = self._values.minKey(key)
            keys.append(own)
        except ValueError:
            pass
        return min(keys)

    def _getField(self, name):
        field = self._fields.get(name, _MARKER)
        if field is _MARKER:
            return self.parents._getField(name)
        if isinstance(field, basestring):
            recordName = field
            while isinstance(field, basestring):
                recordName = field
                field = self._fields[recordName]
            field = FieldRef(recordName, field)
        return field



from .proxy import LineageRegistry
from .proxy import REGISTRY_NAME
from plone.registry.interfaces import IRegistry
from zope.component import getSiteManager


def enableRegistry(child):
    sm = getSiteManager(context=child)
    if REGISTRY_NAME not in child.objectIds():
        child[REGISTRY_NAME] = LineageRegistry(REGISTRY_NAME, parent=child)
    sm.registerUtility(component=child[REGISTRY_NAME], provided=IRegistry)


def disableRegistry(child):
    if REGISTRY_NAME not in child.objectIds():
        return
    # we keep the registry here (intentionally)
    sm = getSiteManager(context=child)
    sm.unregisterUtility(component=child[REGISTRY_NAME], provided=IRegistry)



from .utils import disableRegistry
from .utils import enableRegistry
from collective.lineage.interfaces import IChildSiteCreatedEvent
from collective.lineage.interfaces import IChildSiteRemovedEvent
from zope.component import adapter


@adapter(IChildSiteCreatedEvent)
def enableChildRegistry(event):
    enableRegistry(event.object)


@adapter(IChildSiteRemovedEvent)
def disableChildRegistry(event):
    disableRegistry(event.object)



from collective.lineage.testing import LINEAGE_FIXTURE
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from zope.configuration import xmlconfig


class LineageRegistry(PloneSandboxLayer):

    defaultBases = (LINEAGE_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        import lineage.registry
        xmlconfig.file('configure.zcml', lineage.registry,
                       context=configurationContext)

    def setUpPloneSite(self, portal):
        self['portal'] = portal

LINEAGEREGISTRY_FIXTURE = LineageRegistry()
LINEAGEREGISTRY_INTEGRATION_TESTING = IntegrationTesting(
    bases=(LINEAGEREGISTRY_FIXTURE, ),
    name="lineage.registry:Integration"
)



from setuptools import setup, find_packages
import os

version = '1.2'
shortdesc = u"Lineage AddOn: plone.app.registry for subsites"
longdesc = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
longdesc += open(os.path.join(os.path.dirname(__file__), 'HISTORY.rst')).read()
longdesc += open(os.path.join(os.path.dirname(__file__), 'LICENSE.rst')).read()

setup(
    name='lineage.registry',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    keywords='',
    author='BlueDynamics Alliance',
    author_email='dev@bluedynamics.com',
    url=u'https://github.com/collective/lineage.registry',
    license='GNU General Public Licence',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['lineage'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'collective.lineage',
    ],
    extras_require={
        'test': [
            'collective.lineage [test]',
            'interlude',
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
